-- PROJECT:  MENU.EX;  Menu runs batch files *without* staying in the memory,
--                     the same way it would run by the OS command-line.
--                     * Please read the following DETAILS before use!
-- AUTHOR:   Shian Lee;  http://qe.site90.com
-- VERSION:  1.01   11:09  5/1/2011
-- LANGUAGE: Euphoria programming language version 3.1.1
--           Download free from: http://www.RapidEuphoria.com
--
-- DETAILS:  1. This program is using standard video text mode (80x25).
--           2. This program works only as EXE file (not as source file).
--           3. This program does NOT stay in memory while executing an item.
--           4. This program is using an external editor for editing the menu
--              file. For this I recommend the free QE Editor (download from
--              http://qe.site90.com), or the classic MS-DOS Editor "EDIT"
--              that is part of the DOS and Windows OS (enter 'EDIT' on the
--              command-line - it should be in the PATH...).
--           5. If a menu-item's batch command is calling another batch file,
--              then "CALL" command must be used to finally return to the menu
--              (Example: CALL backup.bat).
--           6. Read the program's HELP_SCREEN for basic usag.
--           7. How to use this program:
--              (1) Compile this source file to create an EXE file;
--                  - Use Euphoria v3.1.1 (for DOS32) or v4.x.x compiler,
--                    download from: http://www.RapidEuphoria.com.
--                  - if using Euphoria v4.x.x for WINDOWS then compile with
--                    the '-con' switch (creates a console program).
--              (2) Place the EXE file in a new directory (Example: C:\MENU),
--                  - It's a good idea to add this directory to the OS PATH.
--              (3) Run the EXE file once to create or update a startup batch
--                  file (the program will display a message).
--              (4) Always run the startup batch file to run this program.
--                  - When executing a menu-item, the program will abort and
--                    return control to the startup batch file;
--                  - The startup batch file will run a temporary batch file
--                    with your batch commands;
--                  - The temporary batch file will finally run the startup
--                    batch file that will run this program again (loop).
--                  * With the above method you run any batch file safely,
--                    exactly the way you would run from the OS command-line,
--                    without losing even one bit of memory!
--           8. This source is coded for DOS32 and WINDOWS compiler; For other
--              OS you may need to change very few lines, specially those
--              lines that create the startup and temporary batch files; Also
--              make sure that you are using the correct slash seperator for
--              your OS.
--              - Check out these:
--                (1) constant SLASH - ('\\' or '/')
--                (2) procedure create_START_file() - (batch commands).
--                (3) procedure create_RUN_file() - (batch commands).
--                (4) procedure abort_program() - (resetting video).
--                (5) module level 'Start The Program - (resetting video).
--           9. Enjoy using this program!
--
-- DISCLAIMER: Use at your own risk. (please report any problem at QE site).
-- ==========================================================================



--bookmark: Programmer Settings;
without type_check  -- speed up include files and program for final EXE.
-- with trace          -- debug the program.



--bookmark: Includes;
include graphics.e      -- text_color()...
include get.e           -- wait_key()...
include wildcard.e      -- upper()...
include misc.e          -- platform()...



-- booleans:
constant
    FALSE = 0, TRUE = 1



-- PURPOSE: test if x is a boolean (0 or 1).
type boolean(object x)
    if integer(x) then
	return x = FALSE or x = TRUE
    else
	return FALSE
    end if
end type



-- PURPOSE: 'if..then..else' as a very useful function.
-- RETURN:  object; ifTrue or ifFalse.
function iif(boolean condition, object ifTrue, object ifFalse)
    if condition then
	return ifTrue
    else
	return ifFalse
    end if
end function



--bookmark: Constants;
constant
    -- program title & version:
    PROGRAM_TITLE = "Menu 1.01",
    -- operating system slash ('\\' for DOS and WINDOWS):
    SLASH = iif(platform() = DOS32 or platform() = WIN32, '\\', '/'),
    -- program's colors (for Color mode only):
    TXT_COLOR = BRIGHT_WHITE,
    BACK_COLOR = BLUE,
    -- default external editor name (must not be longer then EDITOR_LENGTH!):
    DEFAULT_EDITOR = "EDIT",    -- (normally MS-EDIT is in the OS PATH).

    -- maximum input_string length - should be set to 40 characters:
    INPUT_LENGTH = 40,
    -- editor path name maximum length (leave much place for DATA_FILE path):
    EDITOR_LENGTH = INPUT_LENGTH,   -- (INPUT_LENGTH is maximum legal length).
    -- search string maximum length:
    SEARCH_LENGTH = 15,             -- (INPUT_LENGTH is maximum legal length).


    -- 1 = standard output:
    SCREEN = 1,

    -- function keys:
    F1 = 315, F2 = 316, F3 = 317, F4 = 318, F5 = 319,
    F6 = 320, F7 = 321, F8 = 322, F9 = 323, F10 = 324,

    -- other keyboard keys:
    BACKSPACE = 8, ENTER = 13, ESC = 27, SPACE = 32, HOME = 327, UP = 328,
    PAGE_UP = 329, LEFT = 331, RIGHT = 333, END = 335, DOWN = 336,
    PAGE_DOWN = 337, DELETE = 339,

    -- files that this program creates (only names, *without* path):
    CONFIG_FILE = "MENU.INI",   -- initialization file.
    START_FILE = "MN.BAT",      -- startup batch file to start this program.
    RUN_FILE = "~RUNITEM.BAT",  -- temporary batch file that runs menu item.
    DATA_FILE = "MENU.DAT",     -- stores all menu-items data.

    -- total menu items per menu-page:
    PAGE_HEIGHT = 22,

    -- notice: symbols shouldn't be a batch command by mistake...!
    -- string symbol that indicates an item's title in DATA_FILE:
    IS_ITEM = "--",
    LENGTH_IS_ITEM = length(IS_ITEM),
    -- one character symbol (ASCII code) that indicates comment in DATA_FILE:
    IS_COMMENT = '*',

    -- the command line sequence:
    COMMAND = upper(command_line()),
    -- validation switch passed from START_FILE to the program:
    START_SWITCH = "-STARTMENU",

    -- Help screen (23 lines maximum):
    HELP_SCREEN = {
	"1. <F10> = quit program. <Enter> = execute current item.",
	"2. <Up/Down/PgUp/PgDn/Home/End> = scroll menu.",
	"3. Type first letter of an item to quickly jump to it.",
	"4. <F2> = enter search string. <F3> = search again.",
	"5. <F7> = enter external editor name (Example: C:\\DOS\\QBASIC /EDIT).",
	"6. <F4> = edit the menu file with the external editor:",
	"   1) \"" & IS_COMMENT & "\" = Comment line (Example: " & IS_COMMENT & " This line is ignored).",
	"   2) \"" & IS_ITEM & "\" = Item line (Example: " & IS_ITEM & " Norton Commander 5.5).",
	"   3) Lines after \"" & IS_ITEM & "\" are the item's batch commands.",
	"   4) Call other batch files with \"CALL\" command (Example: CALL backup.bat).",
	"   5) Example of the menu file (" & DATA_FILE & "):",
	"       " & IS_ITEM & " Norton Commander 5.5",
	"       C:\\NC\\NC.EXE",
	"       " & IS_ITEM & " Backup my documents",
	"       C:",
	"       CD C:\\DOC",
	"       " & IS_COMMENT & " The CALL command will return control to the Menu program:",
	"       CALL C:\\UTIL\\BACKUP.BAT",
	"7. <F8> = run batch command (Example: CALL backup.bat).",
	"8. Input string keys: <Del> = clear all; <Right> = restore all;",
	"   <Left/Backspace> = delete last character; <Enter> = execute.",
	"9. Remember: run this program with " & START_FILE & ".",
	"    Thank you.   (http://qe.site90.com)"
    }



--bookmark: Variables;
integer
    -- program's colors (for changing B&W and Color mode):
    Txt_Color, Back_Color
    -- initialize colors to display errors before program actually starts:
    Txt_Color = WHITE
    Back_Color = BLACK



-- PURPOSE: set both text and back screen colors at once (in NORMAL_COLOR
--          or REVERSE_COLOR mode).
-- USING:   iif
-- INPUT:   Txt_Color, Back_Color
-- NOTICE:  this routine makes sure that text color range is 0..15 and
--          back color range is 0..7 (this avoids blinking under some OS).
-- SEEALSO: set_program_colors
constant
    NORMAL_COLOR = FALSE, REVERSE_COLOR = TRUE  -- reverseColors constants
procedure set_color(boolean reverseColors)
    text_color(remainder(iif(reverseColors, Back_Color, Txt_Color), 16))
    bk_color(remainder(iif(reverseColors, Txt_Color, Back_Color), 8))
end procedure



-- PURPOSE: hide the cursor, reset color to NORMAL_COLOR and clear the screen.
-- USING:   set_color
-- SEEALSO: draw_menu_screen
procedure clear_the_screen()
    cursor(NO_CURSOR)
    set_color(NORMAL_COLOR)
    clear_screen()
end procedure



-- PURPOSE: clear the keyboard buffer before continue (sometimes, the user's
--          hand might be 'stuck' on a key - this may execute a command...).
-- SEEALSO: pause
procedure clear_keyboard_buffer()
    while get_key() != -1 do
    end while
end procedure



-- PURPOSE: display "Press any key to continue..." and pause execution.
-- USING:   clear_keyboard_buffer
procedure pause()
    puts(SCREEN, "\nPress any key to continue...")
    cursor(THICK_UNDERLINE_CURSOR)
    -- pause:
    clear_keyboard_buffer()
    if atom(wait_key()) then
    end if
    clear_keyboard_buffer()
    -- move cursor to next line:
    puts(SCREEN, '\n')
end procedure



-- PURPOSE: pad the right side of a string with spaces (up to setLength).
-- RETURN:  sequence; a string of length setLength.
-- SEEALSO: trim
function lset(sequence text, integer setLength)
    if length(text) > setLength then
	text = text[1..setLength]
    end if
    return text & repeat(SPACE, setLength - length(text))
end function



-- PURPOSE: remove leading and/or trailing spaces (ASCII-32) and any control
--          characters from a string.
-- DETAILS: side: 0=right side, 1=left side, 2=both sides.
-- RETURN:  sequence; trimed string.
-- SEEALSO: remove_eol, lset
constant
    -- side constants:
    TRIM_RIGHT = 0, TRIM_LEFT = 1, TRIM_BOTH = 2
function trim(sequence text, integer side)
    -- trim right side:
    if side = TRIM_RIGHT or side = TRIM_BOTH then
	while length(text) > 0 and text[$] <= SPACE do
	    text = text[1..$ - 1]
	end while
    end if

    -- trim left side:
    if side = TRIM_LEFT or side = TRIM_BOTH then
	while length(text) > 0 and text[1] <= SPACE do
	    text = text[2..$]
	end while
    end if

    return text
end function



-- PURPOSE: remove \n character from the end of a string (line).
--          useful when reading a file in text mode.
-- RETURN:  sequence; string without \n (End-Of-Line) character.
-- SEEALSO: trim, open_text_file_for_reading
function remove_eol(sequence line)
    if length(line) > 0 then        -- verify that string is not empty.
	if line[$] = '\n' then      -- verify that string has \n character.
	    return line[1..$ - 1]   -- cut end-of-line
	end if
    end if

    return line                     -- do nothing
end function



-- PURPOSE: print static frame, title and status line (function keys).
-- USING:   clear_the_screen, set_color, iif, lset
-- SEEALSO: main_input_loop, print_menu_items
procedure draw_menu_screen()
    sequence func, verticalLines

    -- 10 function keys data:
    func = {
	"Help",     -- F1
	"Search",   -- F2
	"Again",    -- F3
	"Edit",     -- F4
	"Color",    -- F5
	"Clock",    -- F6
	"Editor",   -- F7
	"Run",      -- F8
	"Sleep",    -- F9
	"Quit"      -- F10
    }

    clear_the_screen()  -- also set NORMAL_COLOR

    -- print the frame lines:
    verticalLines = '' & repeat(SPACE, 78) & ''
    position(1, 1)                              -- top horizontal line
    puts(SCREEN, '' & repeat('', 78) & '')
    for row = 2 to 23 do                        -- both vertical lines
	position(row, 1)
	puts(SCREEN, verticalLines)
    end for
    position(24, 1)                             -- bottom horizontal line
    puts(SCREEN, '' & repeat('', 78) & '')

    -- print program's title on top line:
    position(1, 2)
    puts(SCREEN, SPACE & PROGRAM_TITLE & SPACE)

    -- print status line (function keys):
    for i = 0 to 9 do
	-- print function key number in normal colors:
	set_color(NORMAL_COLOR)
	position(25, i * 8 + not(i = 9))
	puts(SCREEN, iif(i = 9, "10", i + 49))
	-- print function key name in reverse colors:
	set_color(REVERSE_COLOR)
	puts(SCREEN, lset(func[i + 1], 6))
    end for

    -- reset color back to normal (after using reverse color):
    set_color(NORMAL_COLOR)
end procedure



-- PURPOSE: print the clock on top of the menu screen.
-- USING:   iif
-- SEEALSO: main_input_loop, hide_clock
procedure clock()
    sequence now
    integer seconds

    now = date()
    seconds = iif(remainder(now[6], 2) = 0, SPACE, ':')

    position(1, 75)
    puts(SCREEN, sprintf("%2d" & seconds & "%02d", {now[4], now[5]}))
end procedure



-- PURPOSE: hide the clock on top of the menu screen.
-- SEEALSO: clock
procedure hide_clock()
    position(1, 70)
    puts(SCREEN, repeat('', 10))
end procedure



-- PURPOSE: print any status message at the bottom-right corner of the menu,
--          (erase old message as well).
-- SEEALSO: print_menu_items
procedure print_status_message(sequence message)
    position(24, 25)    -- column 25 is after the item's serial number status.
    puts(SCREEN, repeat('', 55 - length(message)) & message)
end procedure



-- PURPOSE: print all menu items of the current page.
-- USING:   set_color, iif, lset, print_status_message
-- SEEALSO: main_input_loop, draw_menu_screen
procedure print_menu_items(sequence items, integer startItem,
				    integer currentItem, integer maxItems)

    -- print menu items of the current page:
    for i = startItem to startItem + PAGE_HEIGHT - 1 do
	if i <= maxItems then
	    set_color(iif(i = currentItem, REVERSE_COLOR, NORMAL_COLOR))
	    position(i - startItem + 2, 2)
	    puts(SCREEN, SPACE & lset(items[i], 76) & SPACE)
	end if
    end for

    -- reset color back to normal (after using reverse color):
    set_color(NORMAL_COLOR)

    -- erase any old status message:
    print_status_message("")

    -- print item's serial number on the bottom-left corner:
    position(24, 2)
    puts(SCREEN, sprintf("[%d/%d]", {currentItem, maxItems}) & repeat('', 19))
end procedure



-- PURPOSE: reset video to standard text mode, set B&W colors, clear the
--          screen, show the cursor and abort the program.
-- NOTICE:  DOS32 compiler should do it all when resetting video mode, but if
--          the program is modified/used for other compilers - then this
--          routine will make sure that all is done correctly.
-- SEEALSO: create_START_file, abort_functions, abort_if_open_failed
procedure abort_program()
    -- reset video mode to default text mode (it's enough for DOS compiler):
    if graphics_mode(-1) then
    end if

    -- make sure all is done for other compilers:
    if platform() != DOS32 then
	-- reset colors to standard B&W colors and clear the screen:
	text_color(WHITE)
	bk_color(BLACK)
	clear_screen()
	-- make sure that the cursor is visible:
	cursor(THICK_UNDERLINE_CURSOR)
    end if

    -- abort program with no special error code (no real need for this):
    abort(0)
end procedure



-- PURPOSE: if open file failed then display warning and abort program.
-- DETAILS: fn = file number.
-- USING:   clear_the_screen, pause, abort_program
-- SEEALSO: open_text_file_for_reading
procedure abort_if_open_failed(integer fn, sequence filePathName)
    if fn != -1 then
	return          -- if open successful the do nothing.
    end if

    -- display warning:
    clear_the_screen()
    puts(SCREEN, "\n " & repeat('-', 78) & "\n\n")
    puts(SCREEN, " Warning: can't open the file '" & filePathName & "'!\n\n")
    puts(SCREEN, " Maybe the file is locked, or has a Read-Only attribute.\n")
    puts(SCREEN, " Please fix this problem and try again.\n")
    puts(SCREEN, "\n " & repeat('-', 78) & "\n\n")
    -- pause and abort:
    pause()
    abort_program()
end procedure



-- PURPOSE: create startup batch file that runs this program (START_FILE).
-- DETAILS: - This START_FILE pass the START_SWITCH parameter to the program,
--            to verify that the program executed by the START_FILE;
--          - Executing a menu-item aborts the program and returns control
--            to the START_FILE, that runs the RUN_FILE;
--          - RUN_FILE executes menu-item and runs the START_FILE again (loop).
--          * The above method release *all* memory for other applications.
-- USING:   abort_if_open_failed, pause, abort_program
-- NOTICE:  The START_FILE is created/updated when ever the user executes
--          this program directly (not by the START_FILE).
-- SEEALSO: main_input_loop, create_RUN_file
procedure create_START_file(sequence progDir, sequence progName)
    integer fn

    -- create text file for writing:
    fn = open(progDir & START_FILE, "w")

    -- if open file failed (fn = -1) then display warning and abort program:
    abort_if_open_failed(fn, progDir & START_FILE)

    -- write the START_FILE (OS batch commands):
    puts(fn, "@ECHO OFF\n")
    puts(fn, "REM -- Run this batch file to start '" & progName & "'.\n")
    puts(fn, "REM -- (Run '" & progName & "' directly to update this file).\n")
    -- execute this program (using the special switch):
    puts(fn, progDir & progName & SPACE & START_SWITCH & "\n")
    -- execute a menu-item:
    puts(fn, progDir & RUN_FILE & "\n")
    close(fn)

    -- display message to user that tells to run the START_FILE:
    puts(SCREEN, "\n " & repeat('*', 78))
    puts(SCREEN, "\n " & repeat('*', 78) & "\n\n")
    puts(SCREEN, "  The file '" & progDir & START_FILE & "' was created/updated for you,\n")
    puts(SCREEN, "  Now you may run '" & START_FILE & "' to start this program! Thank you.\n")
    puts(SCREEN, "\n " & repeat('*', 78))
    puts(SCREEN, "\n " & repeat('*', 78) & "\n\n")

    -- pause and abort the program to let user run the START_FILE:
    pause()
    abort_program()
end procedure



-- PURPOSE: create batch file that runs the current menu-item (RUN_FILE).
-- DETAILS: 1. This RUN_FILE executes menu-item's batch commands, then runs
--             the START_FILE to start this program again (loop).
--          2. currentItem > 0 : runs current menu-item batch commands.
--             currentItem = 0 : runs external editor or command.
--             currentItem = -1: quit this program.
--          3. batchLines must include \n character at the end of lines.
-- USING:   abort_if_open_failed
-- NOTICE:  batch command that runs other batch file must use the "CALL"
--          command (example: CALL backup.bat), to return to this program.
-- SEEALSO: abort_functions, create_START_file
procedure create_RUN_file(sequence progDir, integer currentItem,
							sequence batchLines)
    integer fn

    -- create text file for writing:
    fn = open(progDir & RUN_FILE, "w")

    -- if open file failed (fn = -1) then display warning and abort program:
    abort_if_open_failed(fn, progDir & RUN_FILE)

    -- write the RUN_FILE (OS batch commands):
    if currentItem > -1 then    -- (if not quit program)
	puts(fn, "@ECHO OFF\n")
	-- the batch commands (lines already include \n):
	for i = 1 to length(batchLines) do
	    puts(fn, batchLines[i])
	end for
	-- pause after executing batch commands:
	puts(fn, "ECHO *** Press any key to continue...  (Press CTRL+C to stop)\n")
	puts(fn, "PAUSE > NUL\n")
	-- execute the START_FILE (loop back to this program):
	puts(fn, progDir & START_FILE & '\n')
    end if
    close(fn)
end procedure



-- PURPOSE: set the first (top) item in the menu according to currentItem.
-- RETURN:  integer; startItem
-- SEEALSO: set_currentItem
function set_startItem(integer startItem, integer currentItem, integer maxItems)
    -- 1) First: fix startItem according to currentItem:
    if startItem > currentItem then
	startItem = currentItem
    elsif currentItem - startItem >= PAGE_HEIGHT then
	startItem = currentItem - PAGE_HEIGHT + 1
    end if

    -- 2) Then: never leave empty lines in the end (with PgDn or wrong config):
    if maxItems - startItem + 1 < PAGE_HEIGHT then
	startItem = maxItems - PAGE_HEIGHT + 1  -- display full menu page
    end if

    -- 3) Last: fix startItem < 1:
    if startItem < 1 then
	startItem = 1
    end if

    return startItem
end function



-- PURPOSE: select an item in the menu according to user input (inKey).
-- DETAILS: inKey =       user input;
--          items =       all items (strings) in menu;
--          startItem =   the first (top) item in current menu page;
--          currentItem = the currently selected item in menu;
--          maxItems =    total number of items in menu.
-- USING:   set_startItem
-- RETURN:  sequence; {integer startItem, integer currentItem}
-- SEEALSO: search_function, editor_function, run_function, basic_functions,
--          abort_functions
function set_currentItem(integer inKey, sequence items, integer startItem,
				    integer currentItem, integer maxItems)
    integer letter
    boolean foundLetter

    letter = upper(inKey)   -- quick jump to item by its first letter

    -- jump to item 1:
    if inKey = HOME then
	currentItem = 1
	startItem = 1

    -- jump to item maxItems:
    elsif inKey = END then
	currentItem = maxItems
	startItem = maxItems - PAGE_HEIGHT + 1

    -- move one item up:
    elsif inKey = UP then
	if currentItem > 1 then
	    currentItem -= 1
	end if

    -- move one item down:
    elsif inKey = DOWN then
	if currentItem < maxItems then
	    currentItem += 1
	end if

    -- move one page up:
    elsif inKey = PAGE_UP then
	if currentItem - PAGE_HEIGHT >= 1 then
	    currentItem -= PAGE_HEIGHT
	    startItem -= PAGE_HEIGHT
	else
	    currentItem = 1
	    startItem = 1
	end if

    -- move one page down:
    elsif inKey = PAGE_DOWN then
	if currentItem + PAGE_HEIGHT <= maxItems then
	    currentItem += PAGE_HEIGHT
	    startItem += PAGE_HEIGHT
	else
	    currentItem = maxItems
	    startItem = maxItems - PAGE_HEIGHT + 1
	end if

    -- quick jump to item by its first letter (in loop):
    elsif letter >= SPACE and letter <= 255 then
	foundLetter = FALSE
	-- search match from next item till last item:
	for i = currentItem + 1 to maxItems do
	    if letter = upper(items[i][1]) then
		foundLetter = TRUE  -- found match, so don't loop back
		currentItem = i
		exit
	    end if
	end for
	-- search match from first item till previous item:
	for i = 1 to (currentItem - 1) * not foundLetter do
	    if letter = upper(items[i][1]) then
		currentItem = i
		exit
	    end if
	end for
    end if

    -- verify and fix startItem value:
    startItem = set_startItem(startItem, currentItem, maxItems)

    return {startItem, currentItem}
end function



-- PURPOSE: file must be exists to be open for reading, so create the
--          file if not exists, then open it for reading.
-- USING:   abort_if_open_failed
-- RETURN:  integer; fn (file number).
-- SEEALSO: remove_eol, trim
function open_text_file_for_reading(sequence filePathName)
    integer fn

    -- open text file for append, to create the file if not exists:
    fn = open(filePathName, "a")

    -- if open seccessful:
    if fn != -1 then
	close(fn)
	-- open file for reading:
	fn = open(filePathName, "r")
    end if

    -- if open file failed (fn = -1) then display warning and abort program:
    abort_if_open_failed(fn, filePathName)

    -- return the file number:
    return fn
end function



-- PURPOSE: get menu items (only the titles) from DATA_FILE.
-- USING:   open_text_file_for_reading, remove_eol, trim, iif
-- RETURN:  sequence; sequence of strings, each string is a menu item.
-- SEEALSO: main_input_loop, get_item_batch_lines
function get_menu_items(sequence progDir)
    integer fn
    object line
    sequence items

    -- (program will abort if fn = -1)
    fn = open_text_file_for_reading(progDir & DATA_FILE)

    items = {}  -- initialize items to empty sequence.

    -- read all menu items:
    while TRUE do
	-- get the next line from DATA_FILE:
	line = gets(fn)

	-- exit at end of file:
	if atom(line) then
	    exit
	end if

	-- remove \n and trim leading spaces:
	line = trim(remove_eol(line), TRIM_LEFT)

	-- search for the next menu-item-title:
	if length(line) >= LENGTH_IS_ITEM then
	    -- add an item to the menu:
	    if equal(line[1..LENGTH_IS_ITEM], IS_ITEM) then
		-- trim leading and trailing spaces of text only:
		line = trim(line[LENGTH_IS_ITEM + 1..$], TRIM_BOTH)
		-- notice: item may be empty after trim, so add one
		-- space to allow quick jump to this item with space.
		items = append(items, line & iif(length(line) = 0, SPACE, {}))
	    end if
	end if
    end while

    close(fn)

    return items
end function



-- PURPOSE: get current menu item's batch lines from DATA_FILE.
-- USING:   open_text_file_for_reading, trim
-- RETURN:  sequence; sequence of strings, each string is a batch line & \n.
-- SEEALSO: abort_functions, create_RUN_file, get_menu_items
function get_item_batch_lines(sequence progDir, integer currentItem)
    integer fn, itemsCount
    object line
    sequence batchLines
    boolean isItem

    -- (program will abort if fn = -1)
    fn = open_text_file_for_reading(progDir & DATA_FILE)

    -- initialize variables:
    batchLines = {}     -- initialize batch lines to empty sequence.
    itemsCount = 0      -- (search for current menu item in file)

    -- read all menu items:
    while TRUE do
	-- get the next line from DATA_FILE:
	line = gets(fn)

	-- exit at end of file:
	if atom(line) then
	    exit
	end if

	isItem = FALSE                  -- reset flag
	line = trim(line, TRIM_LEFT)    -- trim leading spaces.

	-- search for the next menu-item-title:
	if length(line) > LENGTH_IS_ITEM then   -- (minimum IS_ITEM & \n)
	    isItem = equal(line[1..LENGTH_IS_ITEM], IS_ITEM)
	    itemsCount += isItem        -- count items
	end if

	-- add batch line (including \n) of currentItem:
	if itemsCount = currentItem and not isItem then
	    -- don't add empty lines (program must test if batchLines = {}):
	    -- (notice: line is already trimed left side).
	    if not equal(line, "\n") then   -- if sequence line is not empy
		if length(line) > 0 then    -- if it's not the last line
		    if line[1] != IS_COMMENT then    -- if it's not a comment
			batchLines = append(batchLines, line)
		    end if
		end if
	    end if
	-- done adding all batch lines of currentItem, so exit loop:
	elsif itemsCount > currentItem then
	    exit
	end if
    end while

    close(fn)

    return batchLines
end function



-- PURPOSE: get program's configuration from CONFIG_FILE.
-- USING:   open_text_file_for_reading, remove_eol, trim
-- RETURN:  sequence;
--          {sequence editorName, sequence searchText, sequence runCommand,
--           boolean printClock, boolean colorDisplay,
--           integer startItem, integer currentItem}
-- SEEALSO: main_input_loop, put_configuration
function get_configuration(sequence progDir)
    integer fn
    object line
    sequence configItems

    -- (program will abort if fn = -1)
    fn = open_text_file_for_reading(progDir & CONFIG_FILE)

    -- initialize all configuration items to default:
    configItems = {DEFAULT_EDITOR, "", "", FALSE, FALSE, 1, 1}

    -- read configuration items:
    for i = 1 to length(configItems) do
	-- get the next line from CONFIG_FILE:
	line = gets(fn)

	-- exit at end of file:
	if atom(line) then
	    exit
	end if

	-- first remove \n (end of line):
	line = remove_eol(line)
	-- then trim leading and trailing spaces:
	-- notice: searchText (i = 2) shouldn't be trimed, because search
	--         string may contain spaces...
	if i != 2 then
	    line = trim(line, TRIM_BOTH)
	end if

	-- store configuration item (if empty then use default):
	if length(line) > 0 then
	    if i >= 1 and i <= 3 then       -- get strings
		configItems[i] = line
	    elsif i >= 4 and i <= 5 then    -- get booleans
		configItems[i] = equal(line, "1")
	    elsif i > 5 then                -- get integers
		configItems[i] = value(line)
		configItems[i] = configItems[i][2]
		if not integer(configItems[i]) then
		    configItems[i] = 0
		end if
	    end if
	end if
    end for

    close(fn)

    return configItems
end function



-- PURPOSE: put program's configuration in CONFIG_FILE.
-- DETAILS: configItems =
--          {sequence editorName, sequence searchText, sequence runCommand,
--           boolean printClock, boolean colorDisplay,
--           integer startItem, integer currentItem}
-- USING:   abort_if_open_failed
-- SEEALSO: abort_functions, get_configuration
procedure put_configuration(sequence progDir, sequence configItems)
    integer fn

    fn = open(progDir & CONFIG_FILE, "w")   -- create file for writing

    -- if open file failed (fn = -1) then display warning and abort program:
    abort_if_open_failed(fn, progDir & CONFIG_FILE)

    -- put configuration items in file:
    for i = 1 to 3 do                       -- put strings
	puts(fn, configItems[i] & '\n')
    end for
    for i = 4 to length(configItems) do     -- put booleans/integers
	puts(fn, sprintf("%d\n", configItems[i]))
    end for
    close(fn)
end procedure



-- PURPOSE: set program's colors to B&W or Color display.
-- USING:   iif
-- OUTPUT:  Txt_Color, Back_Color
-- SEEALSO: set_color
procedure set_program_colors(boolean colorDisplay)
    Txt_Color = iif(colorDisplay, TXT_COLOR, WHITE)
    Back_Color = iif(colorDisplay, BACK_COLOR, BLACK)
end procedure



-- PURPOSE: prompt the user to enter a string (text).
-- DETAILS: prompt = a short prompt up to 11 characters (Example: "Search:");
--          text = current editable text;
--          maxLength = maximum text length (not more then INPUT_LENGTH).
-- USING:   hide_clock, print_status_message, lset, clear_keyboard_buffer
-- RETURN:  sequence; {sequence text, integer inKey}
-- SEEALSO: search_function, editor_function, run_function
function input_string(sequence prompt, sequence text, integer maxLength)
    sequence originalText
    integer inKey, pos
    boolean firstInput


    -- initialize:
    originalText = text         -- keep the original text for restore
    pos = 80 - maxLength - 2    -- calculate the start editing position
    firstInput = FALSE          -- still didn't input any key
    hide_clock()                -- hide the clock while editing

    -- print the field's static prompt:
    print_status_message(prompt & " [" & repeat(SPACE, maxLength + 1) & ']')

    -- enter the editing loop:
    while TRUE do
	-- print the edited text:
	cursor(NO_CURSOR)
	position(24, pos)
	puts(SCREEN, lset(text, maxLength + 1))

	-- show cursor at the end of edited text:
	position(24, pos + length(text))
	cursor(THICK_UNDERLINE_CURSOR)

	-- wait for user to press a key:
	inKey = wait_key()

	-- user typed a character:
	if inKey >= SPACE and inKey <= 255 then
	    if not firstInput and inKey != SPACE then
		text = ""   -- first input clears the old string automaticly
	    end if
	    if length(text) < maxLength then
		text &= inKey
	    end if
	-- user delete the whole string:
	elsif inKey = DELETE then
	    text = ""
	-- user delete the last character:
	elsif inKey = BACKSPACE or inKey = LEFT then
	    if length(text) > 0 then
		text = text[1..$ - 1]
	    end if
	-- restore original string if string is empty (or just ignore key):
	elsif inKey = RIGHT then
	    if length(text) = 0 then
		text = originalText
	    end if
	-- user enter the string (return ENTER key):
	elsif inKey = ENTER then
	    print_status_message("")    -- erase edit field.
	    exit
	-- any other key - exit editing and return that key for use:
	else
	    text = originalText         -- restore original string
	    print_status_message("")    -- erase edit field.
	    if inKey = ESC then         -- don't return escape key,
		inKey = -1              --  just let it exit the field.
	    end if
	    exit
	end if

	firstInput = TRUE       -- by now user already input the first key
    end while

    cursor(NO_CURSOR)           -- hide the cursor after editing.
    clear_keyboard_buffer()     -- make sure user is still not pressing a key

    return {text, inKey}
end function



-- PURPOSE: let user search for menu-item by entering text (or search again).
-- USING:   input_string, set_startItem, print_status_message
-- RETURN:  sequence; {sequence searchText, integer startItem,
--                     integer currentItem, boolean printItems,
--                     integer inKey}
-- SEEALSO: editor_function, run_function, set_currentItem, basic_functions,
--          abort_functions
function search_function(sequence searchText, sequence items,
				    integer startItem, integer currentItem,
				    integer maxItems, boolean searchAgain)
    sequence upperText
    integer inKey
    boolean printItems, empty
    object temp

    -- if search again then enter the last search string automaticlly:
    if searchAgain then
	inKey = ENTER
    -- if search new string then prompt user to enter a string:
    else
	temp = input_string("Search:", searchText, SEARCH_LENGTH)
	searchText = temp[1]
	inKey = temp[2]
    end if

    -- still didn't find match (don't print menu items):
    printItems = FALSE

    -- search for matching item:
    if inKey = ENTER then
	inKey = -1  -- enter key shouldn't be used again later - reset it
	empty = (length(searchText) > 0)    -- can't search an empty string
	upperText = upper(searchText)       -- search is not case sensitive

	-- search match from next item till last item:
	for i = currentItem + 1 to maxItems * empty do
	    if match(upperText, upper(items[i])) then
		printItems = TRUE  -- print menu items (& don't loop back)
		currentItem = i
		exit
	    end if
	end for
	-- search match from first item till current item:
	for i = 1 to currentItem * not printItems * empty do
	    if match(upperText, upper(items[i])) then
		printItems = TRUE  -- print menu items
		currentItem = i
		exit
	    end if
	end for

	-- set startItem according to currentItem:
	if printItems then
	    startItem = set_startItem(startItem, currentItem, maxItems)
	-- print status if did not find a match:
	else
	    print_status_message("(MATCH NOT FOUND)")
	end if
    end if

    return {searchText, startItem, currentItem, printItems, inKey}
end function



-- PURPOSE: let user enter a new path name for the external editor.
-- USING:   input_string, trim, iif
-- RETURN:  sequence; {sequence editorName, integer inKey}
-- SEEALSO: search_function, run_function, set_currentItem, basic_functions,
--          abort_functions
function editor_function(sequence editorName)
    integer inKey
    object temp

    -- prompt user to enter a new editor path name:
    temp = input_string("Editor:", editorName, EDITOR_LENGTH)
    temp[1] = trim(temp[1], TRIM_BOTH)  -- trim editorName
    inKey = temp[2]

    -- set new editor only if user didn't enter an empty string:
    editorName = iif(length(temp[1]) = 0 or inKey != ENTER, editorName, temp[1])

    -- (ENTER key shouldn't be used again later - reset it):
    return {editorName, iif(inKey = ENTER, -1, inKey)}
end function



-- PURPOSE: let user enter a command to run (as a single line batch file).
-- USING:   input_string, trim, iif
-- RETURN:  sequence; {sequence runCommand, integer inKey}
-- SEEALSO: abort_functions, search_function, editor_function
function run_function(sequence runCommand)
    integer inKey
    object temp

    -- prompt user to enter a run command:
    temp = input_string("Run:", runCommand, INPUT_LENGTH)
    temp[1] = trim(temp[1], TRIM_BOTH)  -- trim runCommand
    inKey = temp[2]

    -- set new command only if user pressed ENTER:
    runCommand = iif(inKey = ENTER, temp[1], runCommand)

    -- use abort_functions() to run command with F8 (if command != {}):
    if inKey = ENTER and length(runCommand) > 0 then
	inKey = F8
    -- if command is empty then reset ENTER key. Also, don't allow F8 key
    -- (only ENTER key will trigger F8 again to run abort_functions()):
    elsif inKey = ENTER or inKey = F8 then
	inKey = -1
    end if

    return {runCommand, inKey}
end function



-- PURPOSE: run basic functions according to user's input (inKey).
-- USING:   clear_the_screen, pause, set_program_colors, hide_clock,
--          clear_keyboard_buffer
-- RETURN:  sequence; {boolean refreshDisplay,
--                     boolean printClock, boolean colorDisplay}
-- SEEALSO: search_function, editor_function, run_function, set_currentItem,
--          abort_functions
function basic_functions(integer inKey,
				    boolean printClock, boolean colorDisplay)
    boolean refreshDisplay

    -- display help:
    if inKey = F1 then
	clear_the_screen()
	-- print help lines:
	for line = 1 to length(HELP_SCREEN) do
	    position(line, 2)
	    puts(SCREEN, HELP_SCREEN[line])
	end for
	position(24, 1)
	pause()
	refreshDisplay = TRUE

    -- set B&W or Color display:
    elsif inKey = F5 then
	colorDisplay = not colorDisplay
	set_program_colors(colorDisplay)
	refreshDisplay = TRUE

    -- show/hide the clock:
    elsif inKey = F6 then
	printClock = not printClock
	if not printClock then      -- hide the clock
	    hide_clock()
	end if
	refreshDisplay = FALSE      -- no need to refresh display

    -- let the screen rest (sleep):
    elsif inKey = F9 then
	text_color(WHITE)
	bk_color(BLACK)
	clear_screen()
	if atom(wait_key()) then end if     -- (pause)
	refreshDisplay = TRUE

    -- do nothing:
    else
	refreshDisplay = FALSE
    end if

    -- make sure user still not pressing the last key:
    clear_keyboard_buffer()

    return {refreshDisplay, printClock, colorDisplay}
end function



-- PURPOSE: create RUN file, put configuration, and abort program - according
--          to user's input (inKey).
-- DETAILS: configItems =
--          {sequence editorName, sequence searchText, sequence runCommand,
--           boolean printClock, boolean colorDisplay,
--           integer startItem, integer currentItem}
-- USING:   get_item_batch_lines, create_RUN_file, print_status_message,
--          put_configuration, abort_program
-- SEEALSO: search_function, editor_function, run_function, set_currentItem,
--          basic_functions
procedure abort_functions(integer inKey, sequence progDir, sequence configItems)
    sequence editorName, runCommand
    integer currentItem
    object temp

    -- execute the current menu-item:
    if inKey = ENTER then
	currentItem = configItems[7]
	temp = get_item_batch_lines(progDir, currentItem)
	if length(temp) > 0 then
	    create_RUN_file(progDir, currentItem, temp)
	else
	    -- don't run empty batch file (display message):
	    print_status_message("(EMPTY)")
	    return
	end if

    -- edit the menu data file (DATA_FILE) using the external editor:
    elsif inKey = F4 then
	editorName = configItems[1]
	create_RUN_file(progDir, 0, {editorName & SPACE & progDir & DATA_FILE & '\n'})

    -- execute batch command (Run):
    elsif inKey = F8 then
	runCommand = configItems[3]
	create_RUN_file(progDir, 0, {runCommand & '\n'})

    -- quit the program:
    elsif inKey = F10 then
	create_RUN_file(progDir, -1, {})    -- erase previous batch data.

    -- do nothing:
    else
	return
    end if

    -- abort program:
    put_configuration(progDir, configItems) -- save current configuration
    abort_program()
end procedure



-- PURPOSE: this routine runs the whole program; it initializes the
--          program's data, then it starts the program's main input loop.
-- USING:   create_START_file, get_configuration, set_program_colors,
--          get_menu_items, set_startItem, set_currentItem, draw_menu_screen,
--          print_menu_items, clock, search_function, editor_function,
--          run_function, basic_functions, abort_functions
procedure main_input_loop()
    sequence items, progDir, progName, editorName, searchText, runCommand
    integer inKey, startItem, currentItem, maxItems
    object temp
    boolean refreshDisplay, printItems, printClock, colorDisplay


    -- get program's directory and name from command line sequence;
    -- find progName position in the full path (C:\PATH\progName):
    temp = length(COMMAND[2]) - find(SLASH, reverse(COMMAND[2])) + 2
    progDir = COMMAND[2][1..temp - 1]   -- program's directory (end with '\')
    progName = COMMAND[2][temp..$]      -- program's EXE file name


    -- if the user started this program directly, then create/update the
    -- startup batch file (START_FILE), and abort program:
    -- (notice: normally ignore startup file while programming).
    if equal(COMMAND[1], COMMAND[2]) then   -- (not while programming)
	if length(COMMAND) != 3 or          -- (Short-Circuit Evaluation)
	    not equal(upper(COMMAND[3]), upper(START_SWITCH))
	then
	    create_START_file(progDir, progName)    -- abort program.
	end if
    end if


    -- get program's configuration from config file:
    temp = get_configuration(progDir)
    editorName = temp[1]
    searchText = temp[2]
    runCommand = temp[3]
    printClock = temp[4]
    colorDisplay = temp[5]
    startItem = temp[6]     -- (this will be verify in the next lines)
    currentItem = temp[7]   -- (this will be verify in the next lines)
    -- verify and fix editor name length:
    if length(editorName) > EDITOR_LENGTH then
	editorName = editorName[1..EDITOR_LENGTH]
    end if
    -- verify and fix search text length:
    if length(searchText) > SEARCH_LENGTH then
	searchText = searchText[1..SEARCH_LENGTH]
    end if
    -- verify and fix run command text length:
    if length(runCommand) > INPUT_LENGTH then
	runCommand = runCommand[1..INPUT_LENGTH]
    end if
    -- update local variables (Txt_Color, Back_Color):
    set_program_colors(colorDisplay)


    -- get menu items from data file:
    items = get_menu_items(progDir)
    -- if menu is empty then show minimum one empty item:
    if length(items) = 0 then
	items = {"(This menu is empty, please add new items - press F1 for help)."}
    end if
    maxItems = length(items)    -- set total number of menu items.
    -- verify and fix currentItem:
    if currentItem < 1 or currentItem > maxItems then
	currentItem = 1
    end if
    -- verify and fix startItem according to currentItem:
    startItem = set_startItem(startItem, currentItem, maxItems)


    -- initialize main loop variables:
    refreshDisplay = TRUE   -- print the menu screen
    printItems = TRUE       -- print the menu items (override by refreshDisplay)


    -- enter the program's main input loop:
    while TRUE do
	-- refresh the menu screen and/or print the menu items:
	if refreshDisplay or printItems then
	    if refreshDisplay then
		draw_menu_screen()
	    end if
	    print_menu_items(items, startItem, currentItem, maxItems)
	    printItems = FALSE      -- reset flag
	    refreshDisplay = FALSE  -- reset flag
	end if


	-- (1)
	-- get user's input and print the clock:
	if printClock then
	    inKey = -1              -- initialize/reset old user's input
	    while inKey = -1 do
		inKey = get_key()   -- get key and continue
		clock()             -- print the clock
	    end while
	-- just wait for user's input (less work for the CPU...):
	else
	    inKey = wait_key()
	end if


	-- (2)
	-- the next functions must follow user's input, because they
	-- return inKey as well (these are using input_string()):
	while inKey = F2 or inKey = F3 or inKey = F7 or inKey = F8 do
	    -- execute Search function according to inKey:
	    if inKey = F2 or inKey = F3 then
		temp = search_function(searchText, items, startItem,
					currentItem, maxItems, inKey = F3)
		searchText = temp[1]
		startItem = temp[2]
		currentItem = temp[3]
		printItems = temp[4]
		inKey = temp[5]
		if inKey = F2 then
		    inKey = -1  -- exit editing field
		end if

	    -- execute Editor function according to inKey:
	    elsif inKey = F7 then
		temp = editor_function(editorName)
		editorName = temp[1]
		inKey = temp[2]
		if inKey = F7 then
		    inKey = -1  -- exit editing field
		end if

	    -- execute Run function according to inKey:
	    elsif inKey = F8 then
		temp = run_function(runCommand)
		runCommand = temp[1]
		inKey = temp[2] -- (may return F8 to execute abort_functions()).
		if inKey = F8 then
		    exit  -- run abort_functions()
		end if
	    end if
	end while


	-- (3)
	-- if one of the functions above reset inKey to -1 then skip the
	-- following functions:
	-- (notice: the Search function above also sets the printItems flag,
	--          because of this set_currentItem() should be skiped).
	if inKey != -1 then
	    -- select item in menu according to inKey:
	    temp = set_currentItem(inKey, items, startItem, currentItem, maxItems)
	    printItems = (startItem != temp[1] or currentItem != temp[2])
	    startItem = temp[1]
	    currentItem = temp[2]

	    -- execute basic functions according to inKey:
	    temp = basic_functions(inKey, printClock, colorDisplay)
	    refreshDisplay = temp[1]
	    printClock = temp[2]
	    colorDisplay = temp[3]

	    -- create RUN file, put configuration, and abort program
	    -- according to inKey (temp = configItems):
	    temp = {editorName, searchText, runCommand,
		    printClock, colorDisplay,
		    startItem, currentItem
	    }
	    abort_functions(inKey, progDir, temp)
	end if
    end while
end procedure



--bookmark: Start The Program;

    -- reset video to default (I hope it's default) text mode (25 x 80):
    if graphics_mode(-1) then
    end if
    -- try to set rows to 25 (reset video might not set it):
    if text_rows(25) = 25 then
    end if

    -- call the main loop to run this program:
    main_input_loop()

-- End of program.



-- ==========================================================================
--bookmark: Version History;
----------------------------

-- Version 1.01:
----------------
-- Changed trim() function to trim also control characters (as TAB) - not
-- only spaces.

-- Version 1.00:
----------------
-- (First version).

-- ==========================================================================
